/* Hysteresis~ by Rasmus Booberg 2021 */

#include "math.h"
#include <m_pd.h>

static t_class *hysteresis_class;

typedef struct _hysteresis
{
	t_object x_obj;
    	t_pd   	 in2;
	t_float  x_1,x_2,x_3,x_4;
	t_float  y_0,y_1,y_2;

} t_hysteresis;

static void hysteresis_reset(t_hysteresis *x){
	x->x_1 = x->x_2 = x->x_3 = x->x_4 = 0.0;
	x->y_0 = x->y_1 = x->y_2 = 0.0;
}

static void *hysteresis_new(t_symbol *s, int argc , t_atom *argv)
{
    if(argc > 1) post("hysteresis~: extra arguments ignored");
    {
	t_hysteresis *x = (t_hysteresis *)pd_new(hysteresis_class);
	outlet_new(&x->x_obj, &s_signal);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
	inlet_new(&x->x_obj, &x->in2, &s_signal, &s_signal);
	hysteresis_reset(x);
	return (x);
    }
}

static t_float calc_dy( double y, double x, double dx , t_float width , t_float sat )
{
	double dl; double l;
	if(sat < 0.001) sat = 0.001; else if(sat > 0.999){ sat = 0.999;}
	if(width < 0.01) width = 0.01; else if(width > 0.95){ width = 0.95;}
	float drive   = 0.69;

	double k 	= 0.47875;         
	double alpha 	= 0.0016;                 
	double c 	= pow((1-width),0.5)-0.01;
	double nc 	= 1-c;
	double y_s 	= 0.5+(1.0-sat);
	double a 	= y_s/(0.000001+6*drive);
    	double q 	= (x + alpha * y) / a;
    	double coth 	= 1.0 / tanh(q);

    	if(q < 0.001 && q > -0.001){
		l = q / 3.0;
		dl = 1./3.;
	}
    	else {
		l = coth - (1.0 / q);
		dl = (1.0 / (q * q)) - (coth * coth)+1.0;
	
	}

    	double y_diff 	= y_s * l - y;

	double signDX 	= ((dx >= 0.0) - (dx < 0.0));
    	double delta_y 	=  (copysign(1,signDX) == copysign(1,y_diff));

    double kap1 	= nc * delta_y;
    double f1Denom 	= nc * signDX * k - alpha * y_diff;



    	double y_s_oa = y_s / a;
    	double y_s_oa_tc = c * y_s_oa;
    	double y_s_oa_tc_talpha = alpha * y_s_oa_tc;

	double f1 = kap1 * y_diff / f1Denom;
    	double f2 = y_s_oa_tc * dl;
    	double f3 = 1.0 - (y_s_oa_tc_talpha * dl);

    return dx * (f1 + f2) / f3;
}




/*-------------------------PERFORM-------------------------------*/

t_int *hysteresis_perform(t_int *w)
{
    t_hysteresis* x = (t_hysteresis *)(w[1]);
    t_float *in1 = (t_float *)(w[2]);
    t_float *p = (t_float *)(w[3]);
    t_float *k = (t_float *)(w[4]);
    t_float *out = (t_float *)(w[5]);
    int n = (int)(w[6]);

    float x_in;
    float x15;
    float dx1;
    float dx15;
    float dx2;
	t_float	 k1;
	t_float	 k2;
	t_float	 k3;
	t_float	 k4;
        t_float width = *k;
    float x1 = x->x_1;
    float x2 = x->x_2;
    float x3 = x->x_3;
    float x4 = x->x_4;
    float y0 = x->y_0;
    float y1 = x->y_1;
    float y2 = x->y_2;

   while (n--) {


	x_in 	= *in1++;
	*out++  = y2;
dx1 	= x1 - x2;
dx2 	= x2 - x3;
dx15 	= ( dx1 + dx2 ) / 2.;
x15 	= (x_in + 2. * x1 + x2 +( dx1 + dx15 ) / 2.) / 4.;

	k1 = calc_dy(y2     ,x1  ,dx1 ,width,*p);
	k2 = calc_dy(y2+k1/2,x15 ,dx15,width,*p);
	k3 = calc_dy(y2+k2/2,x15 ,dx15,width,*p);
	k4 = calc_dy(y2+k3  ,x2  ,dx2 ,width,*p);


y0 = y2 + k1/6. + k2/3. + k3/3. + k4/6.;
y2 = y1 * 0.92 + ( y0 + y2 ) * 0.04;
y1 = y0;
x4 	= x3*0.9+x4*0.1;
x3	= x2*0.915+x3*0.05+x4*0.035;
x2	= x1*0.935+0.04*x2+0.0175*x3+0.0075*x4;
x1	= x_in*0.942+0.03*x1+0.015*x2+0.008*x3+0.005*x4;

	}
    x->y_0 = y0;
    x->y_1 = y1;
    x->y_2 = y2;

    x->x_1 = x1;
    x->x_2 = x2;
    x->x_3 = x3;
    x->x_4 = x4;

    return (w+7);
}

t_int *hysteresis_perf8(t_int *w)
{
    t_hysteresis* x = (t_hysteresis*) (w[1]);
    t_float *in1 = (t_float *)(w[2]);
    t_float *p   = (t_float *)(w[3]);
    t_float *k  = (t_float *)(w[4]);
    t_float *out = (t_float *)(w[5]);
    int n = (int)(w[6]);

    t_float x1 = x->x_1;
    t_float x2 = x->x_2;
    t_float x3 = x->x_3;
    t_float x4 = x->x_4;
    t_float y0 = x->y_0;
    t_float y1 = x->y_1;
    t_float y2 = x->y_2;


	t_float	 k1;
	t_float	 k2;
	t_float	 k3;
	t_float	 k4;

    t_float x_in;
    t_float x15;
    t_float dx1;
    t_float dx15;
    t_float dx2;
    t_float width = *k;
   while (n--) {

	x_in 	= *in1++;
	*out++ 	= y2;

dx1 	= x1 - x2;
dx2 	= x2 - x3;
dx15 	= ( dx1 + dx2 ) / 2.;
x15 	= (x_in + 2. * x1 + x2 +( dx1 + dx15 ) / 2.) / 4.;

	k1 = calc_dy(y2 ,x1,dx1,width,*p);
	k2 = calc_dy(y2+k1/2,x15,dx15,width,*p);
	k3 = calc_dy(y2+k2/2,x15,dx15,width,*p);
	k4 = calc_dy(y2+k3,x2,dx2,width,*p);

y0 = y2 + k1/6. + k2/3. + k3/3. + k4/6.;

y2 = y1 * 0.92 + ( y0 + y2 ) * 0.04;
y1 = y0;

x4 	= x3*0.9+x4*0.1;
x3	= x2*0.915+x3*0.05+x4*0.035;
x2	= x1*0.935+0.04*x2+0.0175*x3+0.0075*x4;
x1	= x_in*0.942+0.03*x1+0.015*x2+0.008*x3+0.005*x4;
	p++;k++;

	}

    x->y_0 = y0;
    x->y_1 = y1;
    x->y_2 = y1;
    x->x_1 = x1;
    x->x_2 = x2;
    x->x_3 = x3;
    x->x_4 = x4;

    return (w+7);
}

/*-------------------PERFORM END----------------------------*/

void dsp_add_hysteresis(t_hysteresis *x, t_sample *in1, t_sample *in2, t_sample *in3, t_sample *out, int n)
{
    if (n&7)
    	dsp_add(hysteresis_perform, 6,(t_int)x, in1,in2,in3, out, n);
    else
    	dsp_add(hysteresis_perf8, 6,(t_int)x, in1, in2, in3, out, n);
}

static void hysteresis_dsp(t_hysteresis *x, t_signal **sp)
{
    dsp_add_hysteresis(x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[3]->s_vec,sp[0]->s_n);
}

void hysteresis_tilde_setup(void)
{
    hysteresis_class = class_new(gensym("hysteresis~"), (t_newmethod)hysteresis_new, 0,
    	sizeof(t_hysteresis), 0, A_GIMME, 0);
    class_addmethod(hysteresis_class, nullfn, gensym("signal"), 0);
    class_addmethod(hysteresis_class, (t_method)hysteresis_reset, gensym("reset"), 0);
    class_addmethod(hysteresis_class, (t_method)hysteresis_dsp, gensym("dsp"), A_NULL);
}